import asyncio, concurrent.futures
import sys
import oneai

from typing import Awaitable, Dict, Iterable, List, Union
from oneai.classes import Input, Output, Skill
from oneai.segments import *


class Pipeline:
    """
    Language AI pipelines allow invoking and chaining multiple Language Skills to process your input text with a single API call.

    ## Attributes

    `steps: list[Skill]`
        A list of Language Skills to process the input text. The order of the skills in the list determines their input.
    `api_key: str, optional`
        An API key to be used in this pipelines `run` calls. If not provided, the global `oneai.api_key` is used.

    ## Methods

    `run(input, api_key=None) -> Output`
        Runs the pipeline on the input text.
    `run_async(input, api_key=None) -> Awaitable[Output]`
        Runs the pipeline on the input text asynchronously.
    `run_batch(batch, api_key=None) -> Dict[Input, Output]`
        Runs the pipeline on a batch of input texts.
    `run_batch_async(batch, api_key=None) -> Awaitable[Dict[Input, Output]]`
        Runs the pipeline on a batch of input texts asynchronously.

    ## Pipeline Ordering

    The order of the skills in the pipeline determines their input:
    * Skills that are placed after a generator Skill will receive its output as input.
    * If there's no preceding generator Skill, the original input text is used.

    ## Example

    >>> my_text = 'ENTER-YOUR-TEXT-HERE'
    >>> pipeline = oneai.Pipeline(steps=[
    ...     oneai.skills.Topics(),
    ...     oneai.skills.Summarize(min_length=20),
    ...     oneai.skills.Entities()
    ... ])
    >>> output = pipeline.run(my_text)
    >>> output.topics
    [oneai.Label(type=topic, span=[0, 10], name=topic1), ...] # topics from my_text
    >>> output.summary.text
    '...' # summary of my_text
    >>> output.summary.entities
    [oneai.Label(type=entity, span=[0, 10], name=entity1), ...] # entities from the summary
    """

    def __init__(self, steps: List[Skill], api_key: str = None) -> None:
        self.steps = tuple(steps)  # todo: validate (based on input_type)
        self.api_key = api_key

        # split into segments of skills, by where these skills should run (custom skills count as a separate segment)
        self._segments = []
        start = 0
        for i, skill in enumerate(steps):
            if skill.run_custom is not None:
                if i - start:
                    self._segments.append(APISegment(self.steps[start:i]))
                start = i + 1
                self._segments.append(CustomSegment(skill))
        if i + 1 - start:
            self._segments.append(APISegment(self.steps[start : i + 1]))

    def run(
        self, input: Union[str, Input, Iterable[Union[str, Input]]], api_key: str = None
    ) -> Output:
        """
        Runs the pipeline on the input text.

        ## Parameters

        `input: str | Input | Iterable[str | Input]]`
            The input text (or multiple input texts) to be processed.
        `api_key: str, optional`
            An API key to be used in this API call. If not provided, `self.api_key` is used.

        ## Returns

        An `Output` object containing the results of the Skills in the pipeline.

        ## Raises

        `InputError` if the input is is invalid or is of an incompatible type for the pipeline.
        `APIKeyError` if the API key is invalid, expired, or missing quota.
        `ServerError` if an internal server error occured.
        """
        return _async_run_nested(self.run_async(input, api_key))

    async def run_async(
        self, input: Union[str, Input, Iterable[Union[str, Input]]], api_key: str = None
    ) -> Awaitable[Output]:
        """
        Runs the pipeline on the input text asynchronously.

        ## Parameters

        `input: str | Input | Iterable[str | Input]]`
            The input text (or multiple input texts) to be processed.
        `api_key: str, optional`
            An API key to be used in this API call. If not provided, `self.api_key` is used.

        ## Returns

        An Awaitable with an `Output` object containing the results of the Skills in the pipeline.

        ## Raises

        `InputError` if the input is is invalid or is of an incompatible type for the pipeline.
        `APIKeyError` if the API key is invalid, expired, or missing quota.
        `ServerError` if an internal server error occured.
        """
        if isinstance(input, (str, Input)):
            return await process_single_input(
                input,
                self._segments,
                api_key=api_key or self.api_key or oneai.api_key,
            )
        elif isinstance(input, Iterable):
            return await self.run_batch_async(input, api_key)
        else:
            raise TypeError(f"pipeline input must be Input, str or iterable of inputs")

    def run_batch(
        self, batch: Iterable[Union[str, Input]], api_key: str = None
    ) -> Dict[Union[str, Input], Output]:
        """
        Runs the pipeline on a batch of input texts.

        ## Parameters

        `batch: Iterable[str | Input]`
            The input texts to be processed.
        `api_key: str, optional`
            An API key to be used in this API call. If not provided, `self.api_key` is used.

        ## Returns

        A dictionary mapping inputs to the produced `Output` objects, each containing the results of the Skills in the pipeline.

        ## Raises

        `InputError` if the input is is invalid or is of an incompatible type for the pipeline.
        `APIKeyError` if the API key is invalid, expired, or missing quota.
        `ServerError` if an internal server error occured.
        """
        return _async_run_nested(self.run_batch_async(batch, api_key))

    async def run_batch_async(
        self, batch: Iterable[Union[str, Input]], api_key: str = None
    ) -> Awaitable[Dict[Union[str, Input], Output]]:
        """
        Runs the pipeline on a batch of input texts asynchronously.

        ## Parameters

        `batch: Iterable[str | Input]`
            The input texts to be processed.
        `api_key: str, optional`
            An API key to be used in this API call. If not provided, `self.api_key` is used.

        ## Returns

        An Awaitable with a dictionary mapping inputs to the produced `Output` objects, each containing the results of the Skills in the pipeline.

        ## Raises

        `InputError` if the input is is invalid or is of an incompatible type for the pipeline.
        `APIKeyError` if the API key is invalid, expired, or missing quota.
        `ServerError` if an internal server error occured.
        """
        return await process_batch(
            batch, self._segments, api_key=api_key or self.api_key or oneai.api_key
        )

    def __repr__(self) -> str:
        return f"oneai.Pipeline({self.steps})"


# for jupyter environment, to avoid "asyncio.run() cannot be called from a running event loop"
pool = concurrent.futures.ThreadPoolExecutor()
is_36 = sys.version_info[:2] == (3, 6)


def _async_run_nested(coru):
    if is_36:
        loop = asyncio.get_event_loop()
        return loop.run_until_complete(coru)
    else:
        try:
            asyncio.get_running_loop()
            return pool.submit(asyncio.run, coru).result()
        except RuntimeError:
            return asyncio.run(coru)



import aiohttp
import trafilatura as tf
import validators

async def extract_article(input, session: aiohttp.ClientSession):
    text = input.text
    if validators.url(text):
        async with session.get(text) as response:
            text = await response.text()
    return tf.extract(text)



__version__ = "0.3.5"

from typing_extensions import Final
import oneai.skills as skills
from oneai.classes import *
from oneai.pipeline import Pipeline

URL: Final[str] = "https://api.oneai.com/api/v0/pipeline"
"""
Base URL for the pipeline API. Only change if you know what you're doing.
"""

api_key = None
"""
Default api key to use. You can use different keys for different pipelines by setting the `Pipeline.api_key` attribute or passing an `api_key` parameter to `Pipeline.run` calls.
"""

MAX_CONCURRENT_REQUESTS: Final[int] = 2
"""
Max number of allowed concurrent requests to be made by the SDK.
Currently only enforced on `pipeline.run_batch`, other calls may be limited by the API
"""
PRINT_PROGRESS = True
"""
Wether to print progress of batch processing in `Pipeline.run_batch` calls
"""



from base64 import b64encode
from dataclasses import dataclass, field
from io import BytesIO
import json
import os
from typing import BinaryIO, Callable, Iterable, List, Literal, TextIO, Type, Union

import aiohttp

from oneai.exceptions import InputError


@dataclass
class Skill:
    """
    A base class for all Language Skills. Use predefined subclasses of this class, or use this class to define your own Skills.

    A Language Skill is a package of trained NLP models. Skills accept text and respond with processed texts and extracted metadata.

    Process texts with Skills using `Pipeline`s

    ### Skill types
    * Generator Skills (`is_generator=True`) process the input and produce a new text based on it. Examples include `Summarize`, `TranscriptionEnhancer`.
    * Analyzer Skills (`is_generator=False`) scan the input and extract structured data. Examples include `Emotions`, `Topics`.

    ## Attributes

    `api_name: str`
        The name of the Skill in the pipeline API.
    `is_generator: bool`
        Whether the Skill is a generator Skill.
    `skill_params: List[str]`
        Names of the fields of the Skill object that should be passed as parameters to the API.
    `label_type: str`
        If the Skill generates labels, the type name of the label.
    `output_attr: str`
        The attribute name of the Skill's output in the Output object.
    `output_attr1: str`
        Only for Skills with 2 outputs (text / labels)
    `run_custom: (Output, aiohttp.ClientSession) -> str | list[Label] | Output`
        A custom function that will be called locally instead of passing the Skill to the API.
        The `Output` object hold the input text (`Output.text`), and output of the previous Skills in the pipeline.
        Can return a string with generated text (when `is_generator=True`), a list of `Label`s (when `is_generator=False`) or an `Output` object for more complex outputs.
        Can use the `aiohttp.ClientSession` to make HTTP requests.
    """

    api_name: str = ""
    is_generator: bool = False
    _skill_params: List[str] = field(default_factory=list, repr=False, init=False)
    # todo: replace all these w/ an output type object + parse conversations from t. enhancer etc.
    label_type: str = ""
    output_attr: str = ""
    output_attr1: str = field(default="", repr=False)
    run_custom: Callable[['Output', aiohttp.ClientSession], Union[str, 'Labels', 'Output']] = None

    def asdict(self) -> dict:
        return {
            "skill": self.api_name,
            "params": {
                p: self.__getattribute__(p) for p in self._skill_params if self.__getattribute__(p)
            }
        }


def skillclass(
    cls: Type = None,
    api_name: str = "",
    label_type: str = "",
    is_generator: bool = False,
    output_attr: str = "",
    output_attr1: str = "",
    run_custom: Callable[['Output', aiohttp.ClientSession], Union[str, 'Labels', 'Output']] = None 
):
    """
    A decorator for defining a Language Skill class. Decorate subclasses of `Skill` with this decorator to provide default values for instance attributes.
    """

    def wrap(cls) -> cls:
        if not issubclass(cls, Skill):
            print(
                f"warning: class {cls.__name__} decorated with @skillclass does not inherit Skill"
            )

        def __init__(self, *args, **kwargs):
            cls_init(self, *args, **kwargs)
            Skill.__init__(
                self,
                api_name=api_name,
                label_type=label_type,
                is_generator=is_generator,
                output_attr=output_attr,
                output_attr1=output_attr1,
                run_custom=run_custom,
            )
            self._skill_params = [
                a
                for a in self.__dict__
                if not (a in Skill.__dict__ or a == "_skill_params")
            ]

        cls_init = cls.__init__
        cls.__init__ = __init__
        return cls

    return wrap if cls is None else wrap(cls)


class Input:
    """
    A base class for all input texts, allowing structured representations of inputs.

    ## Attributes

    `type: str`
        A type hint for the API, suggesting which models to use when processing the input.

    ## Methods

    `get_text() -> str`
        Returns the input as a string. Not implemented by default.
    `parse(text) -> Input`
        A class method. Parse a string into an instance of the Input class. Not implemented by default.
    """

    def __init__(self, type: str):
        self.type = type

    @classmethod
    def parse(cls, text: str) -> "Input":
        """
        A class method. Parse a string into an instance of the Input class. Not implemented by default.

        ## Parameters

        `text: str`
            The text to parse.

        ## Returns

        The `Input` instance produced from `text`.
        """
        raise NotImplementedError()

    def get_text(self) -> str:
        """
        Returns the input as a string (used by `repr`). Not implemented by default.

        ## Returns

        `str` representation of this `Input` instance.
        """
        raise NotImplementedError()

    def __repr__(self) -> str:
        return self.get_text()


class Document(Input):
    """
    Represents any text that doesn't have a structured format

    ## Attributes

    `text: str`
        The text of the document.

    ## Methods

    `get_text() -> str`
        Returns the text of the document.
    `parse(text) -> Document`
        A class method. Parse a string into a `Document` instance.
    """
    type = 'article'

    def __init__(self, text: str):
        self.text = text

    @classmethod
    def parse(cls, text: str) -> "Document":
        """
        A class method. Parse a string into a `Document` instance.

        ## Parameters

        `text: str`
            The text to parse.

        ## Returns

        The `Document` instance produced from `text`.
        """
        return cls(text)

    def get_text(self) -> str:
        """
        Returns the document as a string.

        ## Returns

        `str` representation of this `Documentation` instance.
        """
        return self.text


@dataclass
class Utterance:
    speaker: str
    utterance: str

    def __repr__(self) -> str:
        return f"\n\t{self.speaker}: {self.utterance}"


class Conversation(Input):
    """
    Represents conversations.

    ## Attributes

    `utterances: List[Utterance]`
        A list of `Utterance` objects, each has `speaker` and `utterance` fields.

    ## Methods

    `get_text() -> str`
        Returns the conversation as a JSON string.
    `parse(text) -> Conversation`
        A class method. Parse a string with a structued conversation format or a conversation JSON string into a `Conversation` instance.
    """

    type = 'conversation'

    def __init__(self, utterances: List[Utterance] = []):
        self.utterances = utterances

    def get_text(self) -> str:
        """
        Returns the conversation as a JSON string.

        ## Returns

        `str` representation of this `Conversation` instance.
        """
        return json.dumps(self.utterances, default=lambda o: o.__dict__)

    @classmethod
    def parse(cls, text: str) -> "Conversation":
        """
        A class method. Parse a string with a structued conversation format or a conversation JSON string into a `Conversation` instance.

        ## Parameters

        `text: str`
            The text to parse.

        ## Returns

        The `Conversation` instance produced from `text`.

        ## Raises

        `ValueError` if `text` is not in a valid conversation format.
        """
        try:  # try to parse as JSON
            js = json.loads(text)
            return cls([Utterance(**utterance) for utterance in js])
        except json.JSONDecodeError:  # if not JSON, assume it's a structured conversation
            from oneai.parsing import parse_conversation

            return parse_conversation(text)

    def __str__(self) -> str:
        return f"oneai.Conversation{repr(self.utterances)}"


class Audio(Input):
    """
    Represents audio inputs.

    ## Attributes

    `data: str`
        A base64-encoded string of the audio data.
    `type: str`
        An input-type hint for the API, either `Conversation.type` or `Document.type`.

    ## Methods

    `get_text() -> str`
        Returns the encoded audio data.
    """

    def __init__(
        self,
        file: Union[str, BinaryIO, bytes],
        extension: str=None,
        type: str=Conversation.type
    ):
        """
        Creates a new `Audio` input instance

        ## Parameters

        `file: Union[str, BinaryIO]`
            The audio file to encode. Either a `str` file name, `BinaryIO` object or `bytes` data.
        `extension: str` (optional)
            The file extension of the audio file. If not provided, it will be guessed from the file name.
            Only `.wav` files are supported at the time.
        `type: str` (optional)
            The input-type hint for the API, either `Conversation.type` or `Document.type`.
        """
        super().__init__(type)

        if isinstance(file, str):
            _, ext = os.path.splitext(file)
            if extension and extension not in ext:
                raise InputError(message='file extension does not match', details=f'file name: {file}, extension: {extension}')
            elif ext in ['.wav']:
                file = open(file, 'rb').read()
            else:
                raise InputError(message=f'unsupported file extension {ext}', details='only .wav files are supported at the time')
        elif isinstance(file, BytesIO):
            file = file.read()
        
        self.content_type = 'audio/wav'
        
        self.data = b64encode(file).decode('utf-8')
        self.encoding = 'base64'
    
    def get_text(self) -> str:
        """
        Returns the encoded audio data.

        ## Returns

        `str` representation of this `Audio` instance.
        """

        return self.data

@dataclass
class Span:
    start: int
    end: int
    section: int = 0
    text: str = None

    @classmethod
    def from_json(cls, objects: List[dict], text: str) -> "List[Span]":
        return [] if not objects else [cls(
            start=object.get("start", None),
            end=object.get("end", None),
            section=object.get("section", None),
            text=text
        ) for object in objects]


@dataclass
class Label:
    """
    Represents a label, marking a part of the input text. Attribute values largely depend on the Skill the labels were produced by.

    ## Attributes

    `type: str`
        Label type, e.g. 'entity', 'topic', 'emotion'.
    `name: str`
        Label class name, e.g. 'PERSON', 'happiness', 'POS'.
    `output_spans: list[Span]`
        The spans in the output text that are marked with the label.
    `input_spans: list[Span]`
        The spans in the input text that are relevant to the label. Only appears if the label was produced by a Skill that supports input spans.
    `span_text: str`
        The text of the label.
    `value: str`
        The value of the label.
    `data: Dict[str, Any]`
        Additional data associated with the label.
    """

    type: str = ""
    name: str = ""
    span: List[int] = field(default_factory=lambda: [0, 0])
    output_spans: List[int] = field(default_factory=list)
    input_spans: List[int] = field(default_factory=list)
    span_text: str = ""
    value: str = ""
    data: dict = field(default_factory=dict)

    @classmethod
    def from_json(cls, object: dict) -> "Label":
        return cls(
            type=object.pop("type", ""),
            name=object.pop("name", ""),
            output_spans=Span.from_json(object.pop("output_spans", []), object.get('span_text', None)),
            input_spans=Span.from_json(object.pop("input_spans", []), object.get('span_text', None)),
            span=object.pop("span", [0, 0]),
            span_text=object.pop("span_text", ""),
            value=object.pop("value", ""),
            data=object.pop("data", {}),
        )

    def __repr__(self) -> str:
        return (
            "oneai.Label("
            + ", ".join(f"{k}={repr(v)}" for k, v in self.__dict__.items() if v)
            + ")"
        )


class Labels(List[Label]):
    """
    Wrapper object for a list of `Label` objects. Provides convenience methods to query labels by attribute.

    ## Methods

    `values() -> list[str]`
        Returns a list of all values of the labels.
    `names() -> list[str]`
        Returns a list of all names of the labels.
    `spans() -> list[list[int]]`
        Returns a list of all spans of the labels.
    `values() -> list[str]`
        Returns a list of all span texts of the labels.
    """
    def values(self):
        return [l.value for l in self]

    def names(self):
        return [l.name for l in self]

    def spans(self):
        return [l.span for l in self]

    def span_texts(self):
        return [l.span_text for l in self]


@dataclass
class Output(Input):
    """
    Represents the output of a pipeline. The structure of the output is dynamic, and corresponds to the Skills used and their order in the pipeline.
    Skill outputs can be accessed as attributes, either with the `api_name` of the corresponding Skill or the `output_attr` field.

    ## Attributes

    `text: str`
        The input text from which this `Output` instance was produced.
    `skills: List[Skill]`
        The Skills used to process `text` and produce this `Output` instance.
    `data: List[List[Label] | Output]`
        Output data produced by `skills`, in the same order. Each element can be:
        * A list of labels, marking data points in `text`, e.g entities.
        * A nested `Output` instance, with a new `text`, e.g summary.
    """

    text: Union[Input, str] = ""
    skills: List[Skill] = field(default_factory=list, repr=False)  # not a dict since Skills are currently mutable & thus unhashable
    data: List[Union[Labels, "Output"]] = field(default_factory=list, repr=False)

    def __getitem__(self, name: str) -> Union[Labels, "Output"]:
        return self.__getattr__(name)

    def __getattr__(self, name: str) -> Union[Labels, "Output"]:
        #####TEMP#HACK########
        if name == "business_entities" and hasattr(self, "labs"):
            return self.__getattr__("labs").business_entities
        ######################
        for i, skill in enumerate(self.skills):
            if (
                (skill.api_name and skill.api_name == name)
                or (skill.output_attr and name in skill.output_attr)
                or (type(skill).__name__ == name)
            ):
                return self.data[i]
        raise AttributeError(f"{name} not found in {self}")

    def __dir__(self) -> Iterable[str]:
        return super().__dir__() + [
            skill.output_attr or skill.api_name for skill in self.skills
        ]

    def get_text(self) -> str:
        return self.text if isinstance(self.text, str) else repr(self.text)

    def add(self, skill: Skill, data: Union['Output', Labels]):
        """
        Add data generated by a `Skill` to the output. Data can be a nested `Output` instance or a list of `Label`s.

        ## Parameters
        
        `skill: Skill`
            The Skill which generated the data.
        `data: Output | list[Label]`
            The data to add to the output.
        """
        self.skills.append(skill)
        self.data.append(data)

    def merge(self, output: "Output") -> "Output":
        """
        Merge data from the given `Output` instance into this one, ignoring input text.

        ## Parameters

        `output: Output`
            The `Output` instance to merge into this one.
        """
        self.skills += output.skills
        self.data += output.data

    def __repr__(self) -> str:
        result = f"oneai.Output(text={repr(self.text)}"
        for i, skill in enumerate(self.skills):
            result += f", {skill.output_attr or skill.api_name}={repr(self.data[i])}"
        return result + ")"



import json
from aiohttp import ClientResponse

# todo: input type validation errors


class OneAIError(Exception):
    """
    A base class for all errors raised by the API.

    ## Attributes

    `status_code: int`
        The API status code of the error.
    `message: str`
        A human-readable message describing the error.
    `details: str`
        A string containing details about the error.
    """

    def __init__(self, status_code: int = 0, message: str = "", details: str = ""):
        self.status_code = status_code
        self.message = message
        self.details = details

    def __str__(self) -> str:
        return (
            f"{type(self).__name__}("
            + (f"status_code={self.status_code}, " if self.status_code else "")
            + (f"message={self.message}" if self.message else "")
            + (f", details={self.details})" if self.details else ")")
        )

    __repr__ = __str__


class InputError(OneAIError):
    """An error raised when the input is invalid or is of an incompatible type for the pipeline."""


class APIKeyError(OneAIError):
    """An error raised when the API key is invalid, expired, or missing quota."""


class ServerError(OneAIError):
    """An error raised when the an internal server error occured."""


errors = {  # map http status codes to OneAIError subclasses
    400: InputError,
    401: APIKeyError,
    403: APIKeyError,
    500: ServerError,
    503: ServerError,
}


async def handle_unsuccessful_response(response: ClientResponse):
    try:
        content = json.loads(await response.content.read())
    except:
        content = {}
    raise errors.get(response.status, ServerError)(
        content.get("status_code", response.status),
        content.get("message", response.reason),
        content.get("details", ""),
    )



import asyncio
from datetime import datetime, timedelta
from inspect import isawaitable
from typing import Awaitable, Dict, Iterable, List, Union

import aiohttp
import oneai

from oneai.api import send_pipeline_request
from oneai.classes import Input, Output, Skill
from oneai.exceptions import OneAIError


class Segment:
    async def run(
        self,
        input: Union[Input, Output, str],
        api_key: str,
        session: aiohttp.ClientSession,
    ) -> Output:
        """
        Run the pipeline segment and attach the output to the input if possible.

        ## Parameters

        `input: Input | Output | str`
            The input to process, either the original input text (`Input` or `str`) or the output of the previous segment (`Output`).
            If the input is an `Output`, the segment output will be attached to the parameter.
        `api_key: str`
            The API key to use for API segments.
        `session: aiohttp.ClientSession`
            Client session to use in segments.

        ## Returns

        `Output`
            The output of the segment.
        """
        raise NotImplementedError


# open a client session and send a request
async def process_single_input(
    input: Union[Input, str], segments: List[Segment], api_key: str
) -> Awaitable[Output]:
    timeout = aiohttp.ClientTimeout(total=6000)
    async with aiohttp.ClientSession(timeout=timeout) as session:
        return await _run_segments(session, input, segments, api_key)


# open a client session with multiple workers and send concurrent requests
async def process_batch(
    batch: Iterable[Union[str, Input]], segments: List[Segment], api_key: str
) -> Awaitable[Dict[Union[str, Input], Output]]:
    iterator = iter(batch)
    results = dict()
    exceptions = 0  # number of exceptions occurred
    time_total = timedelta()  # total time spent on all requests
    # length = len(batch) if hasattr(batch, "__len__") else 0
        
    def next_input():  # distribute batch to workers
        try:
            return next(iterator)
        except StopIteration:
            return None  # we need to break loop for each worker, so we ignore StopIteration

    def print_progress(
        time_delta=timedelta(), start=False, end=False
    ):  # todo progress bar for iterables with __len__
        nonlocal time_total

        def time_format(time: timedelta):
            minutes = f"{time.seconds // 60}m " if time.seconds > 59 else ""
            return minutes + f"{time.seconds % 60}s {time.microseconds // 1000}ms"

        total = len(results)
        time_total += time_delta
        if start:
            print(
                f"Starting processing batch with {oneai.MAX_CONCURRENT_REQUESTS} workers",
                end="\r",
            )
        elif end:
            print(
                "Processed %d inputs - %s/input - %s total - %d successful - %d failed"
                % (
                    total,
                    time_format(time_total / total / oneai.MAX_CONCURRENT_REQUESTS),
                    time_format(time_total / oneai.MAX_CONCURRENT_REQUESTS),
                    total - exceptions,
                    exceptions,
                )
            )
        else:
            print(
                "Input %d - %s/input - %s total - %d successful - %d failed        "
                % (
                    total,
                    time_format(time_delta),
                    time_format(time_total / oneai.MAX_CONCURRENT_REQUESTS),
                    total - exceptions,
                    exceptions,
                ),
                end="\r",
            )

    async def req_worker(session):  # run requests sequentially
        nonlocal exceptions

        time_start = datetime.now()
        input = next_input()
        while input:
            try:
                results[input] = await _run_segments(session, input, segments, api_key)
            except Exception as e:  # todo: break loop for some error types
                print(f"\r\033[KInput {len(results)}:", repr(e))
                results[input] = e
                exceptions += 1

            time_end = datetime.now()
            if oneai.PRINT_PROGRESS:
                print_progress(time_end - time_start)
            time_start = time_end
            input = next_input()

    workers = []
    timeout = aiohttp.ClientTimeout(total=6000)
    async with aiohttp.ClientSession(timeout=timeout) as session:
        for _ in range(oneai.MAX_CONCURRENT_REQUESTS):
            worker = asyncio.create_task(req_worker(session))
            workers.append(worker)
        print_progress(start=True)
        await asyncio.gather(*workers)
        print_progress(end=True)
        return results


async def _run_segments(
    session: aiohttp.ClientSession,
    input: Union[Input, str],
    segments: List[Segment],
    api_key: str,
) -> Awaitable[Output]:
    if not segments:  # no skills
        return Output(input)

    output_top = await segments[0].run(input, api_key, session)
    output = output_top
    for segment in segments[1:]:
        while isinstance(
            output, Output
        ):  # find the last output to serve as the next input
            input = output
            output = output.data[-1] if output.data else None
        output = await segment.run(input, api_key, session)
    return output_top


class CustomSegment(Segment):
    def __init__(self, skill: Skill):
        self.skill = skill

    async def run(
        self,
        input: Union[Input, Output, str],
        api_key: str,
        session: aiohttp.ClientSession,
    ) -> Output:
        if not isinstance(input, Output):
            input = Output(input)
        try:
            output = self.skill.run_custom(input, session)
        except Exception as e:
            raise OneAIError(
                message=f"Error while running Skill {self.skill.api_name}: {e}"
            )
        if isawaitable(output):  # support async custom skills
            output = await output
        if isinstance(
            output, (str, Input)
        ):  # skill returned a string, construct an Output object for next segment
            output = Output(output)
        if self.skill.is_generator or not isinstance(
            output, Output
        ):  # either a new text or labels, set them as an attribute of the existing Output object
            input.add(self.skill, output)
        else:  # this is an edge case which should not happen
            input.merge(
                output
            )  # new_output was produced from same text as next_input, merge them
        return input


class APISegment(Segment):
    def __init__(self, skills: List[Skill]):
        self.skills = skills

    async def run(
        self,
        input: Union[Input, Output, str],
        api_key: str,
        session: aiohttp.ClientSession,
    ) -> Output:
        if isinstance(input, Output):
            output = await send_pipeline_request(
                session, input.text, self.skills, api_key
            )
            input.merge(output)
            output = input
        else:
            output = await send_pipeline_request(session, input, self.skills, api_key)
        return output



import re
import oneai
from oneai.classes import Utterance


def parse_conversation(text: str, strict=False) -> oneai.Conversation:
    # try to parse SRT format (I think this should be done more explicitly)
    srt_regex = re.compile(r'\d+\n\d{1,2}:\d{2}:\d{2}[,.]\d{1,3} --> \d{1,2}:\d{2}:\d{2}[,.]\d{1,3}')
    match = srt_regex.match(text)
    if match:
        data_array = srt_regex.split(text)
        return oneai.Conversation(utterances=[
            Utterance(speaker='SPEAKER', utterance=line.strip().replace('\n', ' '))
            for line in data_array[1:]
        ])

    result = []
    lines = re.split(r"\r?\n", text.strip())
    firstLine = True
    structure = None
    currentLineInfo = None
    waitForTextLine = False
    weak = False
    previousObject = None

    for i, line in enumerate(lines):
        if _isEmptyOrWhitespace(line):
            continue

        if waitForTextLine:
            previousObject["text"] = line.strip()
            previousObject["text_line"] = i
            waitForTextLine = False
            continue

        currentLineInfo = _parseSpeakerLine(line)
        if currentLineInfo is None:
            if firstLine:
                raise ValueError(f"Invalid conversation format at line {i}")
            previousObject["text"] += "\n" + line.strip()
            weak = True
            continue

        if firstLine:
            structure = currentLineInfo
            firstLine = False

        weak |= currentLineInfo["weak"]
        if strict and not _comp(structure, currentLineInfo):
            raise ValueError(
                f"Differing conversation format at line {i}, run with strict=False to ignore"
            )

        if previousObject: result.append(
            Utterance(
                speaker=previousObject["speaker"],
                utterance=previousObject["text"]
            )
        )
        previousObject = {
            "speaker": currentLineInfo["speaker"],
            "text": currentLineInfo["text"],
            "speaker_line": i, # what are these properties for? do I want them in Utterance objects?
            "text_line": i,
            "speaker_length": currentLineInfo["speaker_end"]
        }
        
        waitForTextLine = not bool(currentLineInfo["text"])
    if previousObject and not _isEmptyOrWhitespace(previousObject["text"]):
        result.append(
            Utterance(
                speaker=previousObject["speaker"],
                utterance=previousObject["text"]
            )
        )

    return oneai.Conversation(utterances=result)


def _isEmptyOrWhitespace(text):
    return (not text) or (text.isspace())


def _parseSpeakerLine(text):
    value = {
        "weak": True,
        "preTime": False,
        "speaker": None,
        "speaker_end": None,
        "time": False,
        "separator": False,
        "text": None,
    }

    match = re.match(r"\s*\[?[0-9]+:[0-9]+\]?\s*", text)
    if match:
        text = text[match.end() :]
        value["preTime"] = True
        value["weak"] = False

    # capture STRONG pattern - either timestamp or separator
    matchArea = text[: min(len(text), 35)]
    match = re.search(r"\s{1,10}\[?[0-9]{1,2}:[0-9]{1,2}\]?\s*|\s{0,5}[:|-]", matchArea)

    if not match:  # STRONG not found
        # check if speaker only, in all caps - WEAK PATTERN
        match = re.match(r"[A-Z_-]{3,20}$", matchArea)
        if not match:
            return None  # not a valid format - fail
        else:
            value["speaker"] = match.group(0)
            value["speaker_end"] = len(text)
    else:
        value["weak"] = False
        if match.end() < len(text):
            value["text"] = text[match.end() :].strip()
        value["speaker"] = text[: match.start()]
        value["speaker_end"] = match.end()

        if re.search(r"[:|-]$", match.group(0)):
            value["separator"] = True
        else:
            text = text[match.end() :]
            match = re.match(r"\s*[:|-]", text)
            if match:
                value["separator"] = True
                value["speaker_end"] = match.end()
                value["text"] = (
                    text[match.end() :].strip() if match.end() < len(text) else None
                )
            value["time"] = True
    return value


def _comp(a, b):
    return (
        a["separator"] == b["separator"]
        and a["time"] == b["time"]
        and a["preTime"] == b["preTime"]
        and (a["text"] is None) == (b["text"] is None)
    )



from copy import copy
from typing import Awaitable, List, Union
import aiohttp
import oneai

from oneai.classes import Input, Label, Labels, Output, Skill
from oneai.exceptions import APIKeyError, handle_unsuccessful_response


async def send_pipeline_request(
    session: aiohttp.ClientSession,
    input: Union[Input, str],
    steps: List[Skill],
    api_key: str,
) -> Awaitable[Output]:
    if api_key is None or not api_key:
        raise APIKeyError(60001, 'Missing API key', 'Please provide a valid API key, either by setting the global `oneai.api_key` or passing the `api_key` parameter')

    headers = {"api-key": api_key, "Content-Type": "application/json"}
    request = {
        "text": input if isinstance(input, str) else repr(input),
        "steps": [skill.asdict() for skill in steps],
        "input_type": input.type if isinstance(input, Input) else "article",
    }
    if isinstance(input, Input):
        if hasattr(input, 'content_type'):
            request['content_type'] = input.content_type
        if hasattr(input, 'encoding'):
            request['encoding'] = input.encoding

    async with session.post(oneai.URL, headers=headers, json=request) as response:
        if response.status != 200:
            await handle_unsuccessful_response(response)
        else:
            return build_output(steps, await response.json(), input_type=type(input))

def build_output(skills: List[Skill], raw_output: dict, input_type: type=str) -> Output:
    def get_text(index, input_type):
            # get the input text for this Output object. use index=-1 to get the original input text
            # text can be returned as a simple str or parsed to match a given input type
            text = (
                raw_output["output"][index]["text"]
                if index >= 0
                else raw_output["input_text"]
            )
            # return input_type.parse(text) if issubclass(input_type, Input) else text
            return text

    def split_pipeline(skills: List[Skill], i: int):
        # split pipeline at a generator Skill
        first, second = skills[: i + 1], skills[i + 1 :]
        if hasattr(skills[i], "output_attr1") and skills[i].output_attr1:
            # handle skills that create both text and labels
            clone = copy(skills[i])
            clone.is_generator = False
            clone.output_attr = skills[i].output_attr1
            second = (clone, *second)
        return first, second

    def build_internal(
        output_index: int, skills: List[Skill], input_type: type
    ) -> "Output":
        text = get_text(output_index, input_type)
        labels = [
            Label.from_json(label)
            for label in raw_output["output"][output_index]["labels"]
        ]
        data = []
        for i, skill in enumerate(skills):
            if skill.is_generator:
                skills, next_skills = split_pipeline(skills, i)
                data.append(build_internal(output_index + 1, next_skills, str))
                break
            else:
                data.append(
                    Labels(
                        filter(lambda label: label.type == skill.label_type, labels)
                    )
                )
        return Output(text=text, skills=list(skills), data=data)

    generator = raw_output["output"][0].get("text_generated_by_step_id", 0) - 1
    if generator < 0:
        return build_internal(0, skills, input_type)
    else:
        # edge case- first Skill is a generator, or a generator preceded by Skills that didn't generate output
        # in this case the API will skip these Skills,
        # so we need to create filler objects to match the expected structure
        skills, next_skills = split_pipeline(skills, generator)
        return Output(
            text=get_text(-1, input_type),
            skills=list(skills),
            data=[[]] * generator + [build_internal(0, next_skills, str)],
        )



from dataclasses import dataclass
from warnings import warn
from oneai import scraping

from oneai.classes import Skill, skillclass


@skillclass(
    api_name="enhance",
    is_generator=True,
    label_type="replacement",
    output_attr="enhanced",
    output_attr1="replacements",
)
class TranscriptionEnhancer(Skill):
    """
    Enhances conversations, removing fillers and mistakes. Only works with `Conversation` inputs

    ## Output Attributes

    `enhanced: Output`
        An `Output` object containing the enhanced conversation, and output of the following `Skill`s
    `enhanced.text: str`
        The enhanced conversation
    `enhanced.replacements: list[Label]`
        A list of `Label` objects, representing changes made to the text

    ## Example

    >>> conversation = oneai.Conversation([...])
    >>> pipeline = oneai.Pipeline(steps=[
    ...     oneai.skills.TranscriptionEnhancer()
    ... ])
    >>> output = pipeline.run(conversation)
    >>> output.enhanced
    oneai.Output(text='ENHANCED TRANSCRIPTION', replacements=[...])
    """


@skillclass(
    api_name="summarize",
    is_generator=True,
    label_type="origin",
    output_attr="summary",
    output_attr1="origins",
)
@dataclass
class Summarize(Skill):
    """
    Provides a summary of the input

    ##  Attributes

    `min_length: int`
        Minimal desired length (words) of the summary. Omit for automatic length
    `max_length: int`
        Maximal desired length (words) of the summary. Omit for automatic length
    `find_origins: bool, default=False`
        Whether to generate origins of the summary, pointing to the input text

    ## Output Attributes

    `summary: Output`
        An `Output` object containing the generated summary, and output of the following `Skill`s
    `summary.text: str`
        The generated summary
    `summary.origins: list[Label]`
        A list of `Label` objects, mapping words from the summary to the corresponding words in the input text. Only generated if `find_origins=True`

    ## Example

    >>> pipeline = oneai.Pipeline(steps=[
    ...     oneai.skills.Summarize(min_length=10, find_origins=True)
    ... ])
    >>> output = pipeline.run('YOUR-TEXT')
    >>> output.summary
    oneai.Output(text='SUMMARY', origins=[...])
    """

    min_length: int = 0
    max_length: int = 0
    find_origins: bool = False


@skillclass(api_name="emotions", label_type="emotion")
class Emotions(Skill):
    """
    Detects emotions in the input. Supported emotions: [`happiness`, `sadness`, `fear`, `surprise`, `anger`]

    ## Output Attributes

    `emotions: list[Label]`
        A list of `Label` objects, with spans of the input that were detected as conveying an emotion

    ## Example

    >>> pipeline = oneai.Pipeline(steps=[
    ...     oneai.skills.Emotions()
    ... ])
    >>> output = pipeline.run("I'm not interested. Don't call me again!")
    >>> output.emotions
    [oneai.Label(type=emotion, span=[20, 40], name=anger)]
    """


@skillclass(api_name="entities", label_type="entity")
class Entities(Skill):
    """
    Deprecated- use either `Names` or `Numbers` instead

    ## Output Attributes

    `entities: list[Label]`
        A list of `Label` objects, with detected entities and their type

    ## Example

    >>> pipeline = oneai.Pipeline(steps=[
    ...     oneai.skills.Entities()
    ... ])
    >>> output = pipeline.run('Jim bought shares of Acme Corp.')
    >>> output.entities
    [oneai.Label(type=entity, name=PERSON, span=[0, 3], value=Jim), oneai.Label(type=entity, name=ORG, span=[21, 31], value=Acme Corp.)]
    """
    def __post_init__(self):
        warn("Entities Skill is deprecated- use either `Names` or `Numbers` instead")


@skillclass(api_name="keywords", label_type="keyword")
class Keywords(Skill):
    """
    Detects keywords in the input

    ## Output Attributes

    `keywords: list[Label]`
        A list of `Label` objects, with detected keywords

    ## Example

    >>> pipeline = oneai.Pipeline(steps=[
    ...     oneai.skills.Keywords()
    ... ])
    >>> output = pipeline.run('I love this app, it helps me to organize my tasks')
    >>> output.keywords
    [oneai.Label(type=keyword, span=[12, 15], value=app), oneai.Label(type=keyword, span=[44, 49], value=tasks)]
    """


@skillclass(api_name="highlights", label_type="highlight")
class Highlights(Skill):
    """
    Detects highlights of the input

    ## Output Attributes

    `highlights: list[Label]`
        A list of `Label` objects, with detected highlights

    ## Example

    >>> pipeline = oneai.Pipeline(steps=[
    ...     oneai.skills.Highlights()
    ... ])
    >>> output = pipeline.run('YOUR-TEXT')
    >>> output.keywords
    [oneai.Label(value='A-HIGHLIGHT'), ...]
    """


@skillclass(api_name="sentiments", label_type="sentiment")
class Sentiments(Skill):
    """
    Detects sentiments in the input

    ## Output Attributes

    `sentiments: list[Label]`
        A list of `Label` objects, with spans of the input that were detected as having positive or negative sentiment

    ## Example

    >>> pipeline = oneai.Pipeline(steps=[
    ...     oneai.skills.Sentiments()
    ... ])
    >>> output = pipeline.run('I love this app, it helps me to organize my tasks')
    >>> output.sentiments
    [oneai.Label(type=sentiment, span=[0, 49], value=POS)]
    """


@skillclass(api_name="article-topics", label_type="topic", output_attr="topics")
class Topics(Skill):
    """
    Detects topics of the input

    ## Output Attributes

    `topics: list[Label]`
        A list of `Label` objects, with extracted topics

    ## Example

    >>> pipeline = oneai.Pipeline(steps=[
    ...     oneai.skills.Topics()
    ... ])
    >>> output = pipeline.run('One AI is an NLP platform for developers')
    >>> output.topics
    [oneai.Label(type=topic, value=Machine Learning), oneai.Label(type=topic, value=Artificial Intelligence), ...]
    """


@skillclass(api_name="extract-html", is_generator=True, output_attr="html_article", run_custom=scraping.extract_article)
class HTMLExtractArticle(Skill):
    """
    Extracts the main text content of an HTML page. Accepts URLs or HTML strings

    Using URLs is deprecated, send HTML strings or extract HTML content manually instead (try [BeautifulSoup4](https://pypi.org/project/beautifulsoup4/))

    ## Output

    Main text content from the HTML page

    ## Example

    >>> resp = requests.get('https://oneai.com/about-us/')
    >>> pipeline = oneai.Pipeline(steps=[
    ...     oneai.skills.HTMLExtractArticle()
    ... ])
    >>> output = pipeline.run(resp.text)
    >>> output.html_article.text[:100] + '...'
    One AI - About us
    ABOUT ONE AI
    // Bringing human-level language AI to everyday life, one developer a...
    """


@skillclass(api_name="html-extract-text", is_generator=True, output_attr="html_text")
class HTMLExtractText(Skill):
    """
    Extracts all text content of an HTML page (including text from control elements). Accepts URLs or HTML strings. Use HTMLExtractArticle to extract the main content only

    Using URLs is deprecated, send HTML strings or extract HTML content manually instead (try [BeautifulSoup4](https://pypi.org/project/beautifulsoup4/))

    ## Output

    Text content from the HTML page

    ## Example

    >>> resp = requests.get('https://oneai.com/about-us/')
    >>> pipeline = oneai.Pipeline(steps=[
    ...     oneai.skills.HTMLExtractText()
    ... ])
    >>> output = pipeline.run(resp.text)
    >>> output.html_text.text[:100] + '...'
    One AI - About usSkip to main contentAPIStudioSkillsIndustriesPricingResearchAboutAbout UsLeadership...
    """


@skillclass(
    api_name="business-entities",
    is_generator=True,
    label_type="business-entity",
    output_attr="labs",
    output_attr1="business_entities",
)
class BusinessEntities(Skill):
    """
    ### 'Labs' Skill- this Skill is still in beta and is may produce incorrect results in some cases

    deprecated- use either `Pricings` instead.
    Detects business entities (quantities, pricings) in the input

    ## Output Attributes

    `business_entities: list[Label]`
        A list of `Label` objects, with detected business entities and their data

    ## Example

    >>> pipeline = oneai.Pipeline(steps=[
    ...     oneai.skills.BusinessEntities()
    ... ])
    >>> output = pipeline.run('it costs 10 USD per barrel')
    >>> output.business_entities
    [oneai.Label(type=business-entity, name=pricing, data={'amount': 10.0, 'currency': 'USD', 'unit': 'barrel'}, span=[9, 26], value=10 USD per barrel)]
    """
    def __post_init__(self):
        warn("BusinessEntities Skill is deprecated- use either `Pricings` instead")


@skillclass(
    api_name="action-items", label_type="action-item", output_attr="action_items"
)
class ActionItems(Skill):
    """
    ### 'Labs' Skill- this Skill is still in beta and is may produce incorrect results in some cases

    Detects action items in the input

    ## Output Attributes

    `action_items: list[Label]`
        A list of `Label` objects, with detected action items

    ## Example

    >>> pipeline = oneai.Pipeline(steps=[
    ...     oneai.skills.ActionItems()
    ... ])
    >>> output = pipeline.run("It's been a blast talking to you. Let's schedule another meeting for next Sun.")
    >>> output.action_items
    [oneai.Label(type=action-item, span=[34, 78], value=Let's schedule another meeting for next Sun.)]
    """


@skillclass(
    api_name="anonymize",
    is_generator=True,
    label_type="anonymized",
    output_attr="anonymized",
    output_attr1="anonymizations",
)
class Anonymize(Skill):
    """
    Anonymize texts, removing personal information

    ## Output Attributes

    `anonymized: Output`
        An `Output` object containing the anonymized text, and output of the following `Skill`s
    `anonymized.text: str`
        The anonymized text
    `anonymized.anonymizations: list[Label]`
        A list of `Label` objects, representing anonymized information

    ## Example

    >>> pipeline = oneai.Pipeline(steps=[
    ...     oneai.skills.Anonymize()
    ... ])
    >>> output = pipeline.run("hello, I'm Michael, my email is michael@abcde.com")
    >>> output.anonymized.text
    hello, I'm ***, my email is ***
    >>> output.anonymized.anonymizations
    [oneai.Label(type=anonymized, span=[11, 14], value=Michael), oneai.Label(type=anonymized, span=[28, 31], value=michael@abcde.com)]
    """

@skillclass(
    api_name="business-entities",
    is_generator=True,
    label_type="business-entity",
    output_attr="labs",
    output_attr1="pricings",
)
class Pricings(Skill):
    """
    ### 'Labs' Skill- this Skill is still in beta and is may produce incorrect results in some cases

    Detects pricings and quantities in the input

    ## Output Attributes

    `pricings: list[Label]`
        A list of `Label` objects, with detected pricings and their data

    ## Example

    >>> pipeline = oneai.Pipeline(steps=[
    ...     oneai.skills.Pricings()
    ... ])
    >>> output = pipeline.run('it costs 10 USD per barrel')
    >>> output.pricings
    [oneai.Label(name=pricing, data={'amount': 10.0, 'currency': 'USD', 'unit': 'barrel'}, span=[9, 26], value=10 USD per barrel)]
    """


@skillclass(api_name="names", label_type="name")
class Names(Skill):
    """
    Detects and classifies names in the input

    ## Output Attributes

    `names: list[Label]`
        A list of `Label` objects, with detected names and their type

    ## Example

    >>> pipeline = oneai.Pipeline(steps=[
    ...     oneai.skills.Names()
    ... ])
    >>> output = pipeline.run('Jim bought shares of Acme Corp.')
    >>> output.names
    [oneai.Label(type=entity, name=PERSON, span=[0, 3], value=Jim), oneai.Label(type=entity, name=ORG, span=[21, 31], value=Acme Corp.)]
    """


@skillclass(api_name="numbers", label_type="number")
class Numbers(Skill):
    """
    Detects and classifies numbers and dates in the input

    ## Output Attributes

    `entities: list[Label]`
        A list of `Label` objects, with detected numbers and their type

    ## Example

    >>> pipeline = oneai.Pipeline(steps=[
    ...     oneai.skills.Numbers()
    ... ])
    >>> output = pipeline.run('Jim bought shares of Acme Corp.')
    >>> output.numbers
    """


@skillclass(api_name="sentences", label_type="sentence")
class SplitBySentence(Skill):
    """
    Splits input by sentence

    ## Output Attributes

    `sentences: list[Label]`
        A list of `Label` objects, with detected numbers and their type

    ## Example

    >>> pipeline = oneai.Pipeline(steps=[
    ...     oneai.skills.Sentences()
    ... ])
    >>> output = pipeline.run('...')
    >>> output.sentences
    """


@skillclass(api_name="dialogue-segmentation", label_type="dialogue-segment", output_attr="segments")
class SplitByTopic(Skill):
    """
    Splits input by discussed topics

    ## Output Attributes

    `segments: list[Label]`
        A list of `Label` objects, with generated topic-segments

    ## Example

    >>> pipeline = oneai.Pipeline(steps=[
    ...     oneai.skills.SplitByTopic()
    ... ])
    >>> output = pipeline.run('...')
    >>> output.segments
    """


@skillclass(api_name="sales-insights", label_type="sales-insights", output_attr="sales_insights")
class SalesInsights(Skill):
    """
    Splits input by discussed topics

    ## Output Attributes

    `sales_insights: list[Label]`
        A list of `Label` objects, with detected sales insights

    ## Example

    >>> pipeline = oneai.Pipeline(steps=[
    ...     oneai.skills.SalesInsights()
    ... ])
    >>> output = pipeline.run('...')
    >>> output.sales_insights
    """


